/*
 * png.c
 *
 * crafted - a pud editor for the freecraft project.
 * 
 * Copyright (C) 2001-2002 DindinX <David@dindinx.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <png.h>

#include "iolib.h"
#include "crafted.h"
#include "crafted_png.h"

typedef struct Palette
{
  guchar r, g, b;
} Palette;

/**
**	png read callback for CL-IO.
**
**	@param png_ptr	png struct pointer.
**	@param data	byte address to read to.
**	@param length	number of bytes to read.
*/
void CL_png_readfn(png_structp png_ptr,png_bytep data,png_size_t length)
{
  png_size_t check;

  check = (png_size_t)CLread((CLFile *)png_get_io_ptr(png_ptr), data, (size_t)length);
  if (check != length)
  {
    png_error(png_ptr, "Read Error");
  }
}

/**
**	Load a png graphic file.
**
**	FIXME: must support other formats than 8bit indexed
**
**	@param name	png filename to load.
**
**	@return		graphic object with loaded graphic, or NULL if failure.
*/
Graphic *LoadGraphicPNG(const gchar* name)
{
  Graphic      *graphic;
  Palette      *palette;
  CLFile       *fp;
  png_structp   png_ptr;
  png_infop     info_ptr;
  guchar      **lines;
  guchar       *data;
  guint         i;
  guchar       *rgb_buf;
  GdkGC        *gc;
  GdkColor      mask_pattern;

  /*
   * open + prepare
   */
  if (!(fp=CLopen(name)))
  {
    gchar *error_txt;

    error_txt = g_strdup_printf("Can't open file: %s", name);
    perror(error_txt);
    g_free(error_txt);
    return NULL;
  }

  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
  {
    CLclose(fp);
    return NULL;
  }
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
  {
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    CLclose(fp);
    return NULL;
  }
  if (setjmp(png_ptr->jmpbuf))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    CLclose(fp);
    return NULL;
  }
  /* png_init_io(png_ptr,fp); */
  png_set_read_fn(png_ptr, fp, CL_png_readfn);

  /*
   * Setup ready, read header info.
   */
  png_read_info(png_ptr, info_ptr);
  if (verbosity)
  {
    g_message("%s: width %ld, height %ld (%ld bytes)\n",
              name, info_ptr->width, info_ptr->height,
              info_ptr->width*info_ptr->height);
    g_message("%s: %s", name,
              png_get_valid(png_ptr, info_ptr, PNG_INFO_PLTE) ? "palette" : "");
    g_message(" %s", png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS) ? "transparent" : "");
    g_message(" depth %d\n", info_ptr->bit_depth);
  }

  /*
   * Setup translators:
   */
  palette = g_new(Palette, 256);

  if (info_ptr->color_type==PNG_COLOR_TYPE_PALETTE)
  {
    if (verbosity)
      g_message("Color palette\n");
    if (info_ptr->valid & PNG_INFO_PLTE)
    {
      if (verbosity)
        g_message(" palette %d\n", info_ptr->num_palette);
      if (info_ptr->num_palette>256)
      {
        abort();
      }
      for (i=0 ; i<info_ptr->num_palette ; ++i)
      {
        palette[i].r = info_ptr->palette[i].red;
        palette[i].g = info_ptr->palette[i].green;
        palette[i].b = info_ptr->palette[i].blue;
      }
      for ( ; i<256; ++i)
      {
        palette[i].r = palette[i].g = palette[i].b = 0;
      }
    }
  }

  if (info_ptr->bit_depth == 16)
  {
    png_set_strip_16(png_ptr);
  }
  if (info_ptr->bit_depth<8)
  {
    png_set_packing(png_ptr);
  }

  png_read_update_info(png_ptr,info_ptr);

  /*
   * Allocate and reserve memory.
   */
  graphic = g_new(Graphic, 1);
  graphic->width = info_ptr->width;
  graphic->height = info_ptr->height;
  if (info_ptr->width != info_ptr->rowbytes)
  {
    if (verbosity)
      g_message("width(%ld)!=rowbytes(%ld) in file:%s\n",
                info_ptr->width, info_ptr->rowbytes, name);
    abort();
  }

  lines = alloca(graphic->height * sizeof(*lines));
  if (!lines)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    CLclose(fp);
    return NULL;
  }
  data = g_malloc(graphic->height * graphic->width);
  if (!data)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    CLclose(fp);
    return NULL;
  }

  for (i=0; i < graphic->height ; ++i)
  { /* start of lines */
    lines[i] = data+i*graphic->width;
  }

  /*
   * Final read the image.
   */
  png_read_image(png_ptr, lines);
  png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
  CLclose(fp);
  /* Build the pixmap and mask */
  graphic->pixmap = gdk_pixmap_new(drawing_area->window,
                                   graphic->width,
                                   graphic->height, -1);
  graphic->mask = gdk_pixmap_new(drawing_area->window,
                                 graphic->width,
                                 graphic->height, 1);

  gc = gdk_gc_new (graphic->mask);
  mask_pattern.pixel = 0;
  gdk_gc_set_foreground(gc, &mask_pattern);
  gdk_draw_rectangle(graphic->mask, gc, TRUE, 0, 0, -1, -1);
  mask_pattern.pixel = 1;
  gdk_gc_set_foreground(gc, &mask_pattern);

  rgb_buf = g_new(guchar, 3 * graphic->width * graphic->height);
  for (i = 0 ; i<graphic->width * graphic->height ; i++)
  {
    rgb_buf[i*3+0] = palette[data[i]].r;
    rgb_buf[i*3+1] = palette[data[i]].g;
    rgb_buf[i*3+2] = palette[data[i]].b;
    if (data[i] != 255)
    {
      gdk_draw_point(graphic->mask, gc,
                     i % graphic->width,
                     i / graphic->width);
    }
  }
  gdk_draw_rgb_image(graphic->pixmap, drawing_area->style->white_gc,
                     0,0, graphic->width, graphic->height,
                     GDK_RGB_DITHER_NONE, rgb_buf,
                     3*graphic->width);
  g_free(rgb_buf);
  g_free(data);
  g_free(palette);
  return graphic;
}

Graphic *load_graphic_PNG_with_size(const gchar* name, guint width, guint height)
{
  Graphic      *graphic;
  Palette      *palette;
  CLFile       *fp;
  png_structp   png_ptr;
  png_infop     info_ptr;
  guchar      **lines;
  guchar       *data;
  guint         i, j;
  guchar       *rgb_buf;
  GdkGC        *gc;
  GdkColor      mask_pattern;

  /*
   * open + prepare
   */
  if (!(fp=CLopen(name)))
  {
    gchar *error_txt;

    error_txt = g_strdup_printf("Can't open file: %s", name);
    perror(error_txt);
    g_free(error_txt);
    return NULL;
  }

  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
  {
    CLclose(fp);
    return NULL;
  }
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
  {
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    CLclose(fp);
    return NULL;
  }
  if (setjmp(png_ptr->jmpbuf))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    CLclose(fp);
    return NULL;
  }
  /* png_init_io(png_ptr,fp); */
  png_set_read_fn(png_ptr, fp, CL_png_readfn);

  /*
   * Setup ready, read header info.
   */
  png_read_info(png_ptr, info_ptr);
  if (verbosity)
  {
    g_message("%s: width %ld, height %ld (%ld bytes)\n",
              name, info_ptr->width, info_ptr->height,
              info_ptr->width*info_ptr->height);
    g_message("%s: %s", name,
              png_get_valid(png_ptr, info_ptr, PNG_INFO_PLTE) ? "palette" : "");
    g_message(" %s", png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS) ? "transparent" : "");
    g_message(" depth %d\n", info_ptr->bit_depth);
  }

  /*
   * Setup translators:
   */
  palette = g_new(Palette, 256);

  if (info_ptr->color_type==PNG_COLOR_TYPE_PALETTE)
  {
    if (verbosity)
      g_message("Color palette\n");
    if (info_ptr->valid & PNG_INFO_PLTE)
    {
      if (verbosity)
        g_message(" palette %d\n", info_ptr->num_palette);
      if (info_ptr->num_palette>256)
      {
        abort();
      }
      for (i=0 ; i<info_ptr->num_palette ; ++i)
      {
        palette[i].r = info_ptr->palette[i].red;
        palette[i].g = info_ptr->palette[i].green;
        palette[i].b = info_ptr->palette[i].blue;
      }
      for ( ; i<256; ++i)
      {
        palette[i].r = palette[i].g = palette[i].b = 0;
      }
    }
  }

  if (info_ptr->bit_depth == 16)
  {
    png_set_strip_16(png_ptr);
  }
  if (info_ptr->bit_depth<8)
  {
    png_set_packing(png_ptr);
  }

  png_read_update_info(png_ptr,info_ptr);

  /*
   * Allocate and reserve memory.
   */
  graphic = g_new(Graphic, 1);
  graphic->width = info_ptr->width;
  graphic->height = info_ptr->height;
  if (info_ptr->width != info_ptr->rowbytes)
  {
    if (verbosity)
      g_message("width(%ld)!=rowbytes(%ld) in file:%s\n",
                info_ptr->width, info_ptr->rowbytes, name);
    abort();
  }

  lines = alloca(graphic->height * sizeof(*lines));
  if (!lines)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    CLclose(fp);
    return NULL;
  }
  data = g_malloc(graphic->height * graphic->width);
  if (!data)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    CLclose(fp);
    return NULL;
  }

  for (i=0; i < graphic->height ; ++i)
  { /* start of lines */
    lines[i] = data+i*graphic->width;
  }

  /*
   * Final read the image.
   */
  png_read_image(png_ptr, lines);
  png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
  CLclose(fp);
  /* Build the pixmap and mask */
  graphic->pixmap = gdk_pixmap_new(drawing_area->window,
                                   width, height, -1);
  graphic->mask = gdk_pixmap_new(drawing_area->window,
                                 width, height, 1);

  gc = gdk_gc_new (graphic->mask);
  mask_pattern.pixel = 0;
  gdk_gc_set_foreground(gc, &mask_pattern);
  gdk_draw_rectangle(graphic->mask, gc, TRUE, 0, 0, -1, -1);
  mask_pattern.pixel = 1;
  gdk_gc_set_foreground(gc, &mask_pattern);

  rgb_buf = g_new(guchar, 3 * width * height);
  for (j = 0 ; j < height ; j++)
  for (i = 0 ; i < width  ; i++)
  {
    rgb_buf[(j*width+i)*3+0] = palette[data[(j*graphic->width+i)]].r;
    rgb_buf[(j*width+i)*3+1] = palette[data[(j*graphic->width+i)]].g;
    rgb_buf[(j*width+i)*3+2] = palette[data[(j*graphic->width+i)]].b;
    if (data[j*graphic->width+i] != 255)
      gdk_draw_point(graphic->mask, gc, i,j);
  }
  gdk_draw_rgb_image(graphic->pixmap, drawing_area->style->white_gc,
                     0,0, width, height,
                     GDK_RGB_DITHER_NONE, rgb_buf, 3*width);
  gdk_gc_destroy(gc);
  g_free(rgb_buf);
  g_free(data);
  g_free(palette);
  return graphic;
}
